import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

def is_proper_coloring(colors, G):
    assert len(colors) == 4
    for node, other_node in G.edges:
        if colors[node] == colors[other_node]:
            return False
    return True

G = nx.Graph()
G.add_edge(1, 2)
G.add_edge(2, 3)
G.add_edge(3, 4)
G.add_edge(4, 1)
available_colors = ('blue', 'red', 'green', 'yellow')


pos = nx.circular_layout(G)

# +
current_colors = {1: 'blue', 2: 'red', 3: 'yellow', 4: 'red'}
nx.draw(G, pos=pos, node_color=[current_colors[n] for n in G.nodes], with_labels=True)

is_proper_coloring(current_colors, G)
# -

# # Glauber dynamics with uniform distribution on proper colorings as the stationary distribution



states_visited = []
T_steps = 100000
nx.draw(G, pos=pos, node_color=[current_colors[n] for n in G.nodes], with_labels=True)
plt.show()
for t in range(T_steps):
    if (t & (t-1) == 0): print(t, end="...")
    # base chain proposal
    n_t = np.random.choice(G.nodes)
    c_t = np.random.choice(available_colors)
    # proposal from the base chain
    proposal = {**current_colors, n_t: c_t}
    # reject/accept the move based on the metropolis rule
    if is_proper_coloring(proposal, G):
        # accept the move
        current_colors = proposal
        print("accept the move")
    else:
        # reject the move
        current_colors = current_colors
        print("reject the move")
    states_visited.append(tuple(current_colors[n] for n in G.nodes))
    if t< 10:
        print('random node', n_t)
        print('random color', c_t)
        nx.draw(G, pos=pos, node_color=[current_colors[n] for n in G.nodes], with_labels=True)
        plt.show()

from collections import Counter
counts = Counter(states_visited)
counts

# proportion of time points spent at each coloring
pi_estimated = np.array(list(counts.values()))/T_steps
pi_estimated

plt.bar([a for a in range(len(pi_estimated))], pi_estimated)

for coloring in [0,1,2,3,4,5]:
    print(pi_estimated[coloring]**-1)

np.average(pi_estimated)**-1

# # Compare with the exact formula:
# https://math.stackexchange.com/a/3393053/484640
r = len(available_colors)
r*(r-1)*(r**2 - 3*r + 3)


